package com.aitrainer.api.controller

import com.aitrainer.api.model.*
import com.aitrainer.api.service.ServiceBeans
import com.aitrainer.api.repository.CustomerRepository
import com.aitrainer.api.repository.MembershipRepository
import com.aitrainer.api.service.Email
import com.aitrainer.api.service.EmailTemplateService
import com.aitrainer.api.service.Firebase
import org.slf4j.LoggerFactory
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.beans.factory.annotation.Value
import org.springframework.data.repository.findByIdOrNull
import org.springframework.http.HttpHeaders
import org.springframework.http.ResponseEntity
import org.springframework.security.access.annotation.Secured
import org.springframework.web.bind.annotation.*
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter
import java.util.*
import javax.validation.Valid
import java.util.Base64


@RestController
@RequestMapping("/api")
class CustomerController (private val customerRepository: CustomerRepository, private val membershipRepository: MembershipRepository,) {
    private val logger = LoggerFactory.getLogger(javaClass)

    @Autowired
    var serviceBeans: ServiceBeans? = null

    @Autowired
    private var emailTemplateService: EmailTemplateService? = null

    @Secured
    @GetMapping("/customers")
    fun getAllCustomers(@RequestHeader headers: HttpHeaders): List<Customer> =
            customerRepository.findAll()

    @Secured
    @PostMapping("/customers")
    fun createNewCustomer(@Valid @RequestBody customer: Customer, @RequestHeader headers: HttpHeaders): Customer =
            customerRepository.save(customer)

    @Secured
    @GetMapping("/customers/{id}")
    fun getCustomerById(@PathVariable(value = "id") customerId: Long): ResponseEntity<Customer> {
        val customer: Customer? = customerRepository.findById(customerId).orElse(null)
        return if (customer == null) ResponseEntity.notFound().build() else ResponseEntity.ok().body(customer)
    }

    @Secured
    @GetMapping("/customers/find_by_firebaseuid/{uid}")
    fun getCustomerByFirebaseUid(@PathVariable(value = "uid") firebaseUid: String): ResponseEntity<Customer> {
        val customer: Customer? = customerRepository.findByFirebaseUid(firebaseUid)

        return if (customer == null) ResponseEntity.notFound().build() else ResponseEntity.ok().body(customer)
    }

    @Secured
    @GetMapping("/customers/find_by_email/{email}")
    fun getCustomerByEmail(@PathVariable(value = "email") email: String): ResponseEntity<Customer> {
        val customer: Customer? = customerRepository.findByEmailAndActive(email, "Y")
        return if (customer == null) ResponseEntity.notFound().build() else ResponseEntity.ok().body(customer)
    }


    @Secured
    @GetMapping("/customers/real")
    fun getRealCustomers(active: String, @RequestHeader headers: HttpHeaders): ResponseEntity<List<Customer>> =
            ResponseEntity.ok().body(customerRepository.findByActive(active))

    @Secured
    @GetMapping("/customers/trainees/{id}")
    fun getCustomerListByTrainerId(@PathVariable(value = "id") trainerId: Long, @RequestHeader headers: HttpHeaders):
            ResponseEntity<List<Customer>> {
        val customerList: List<Customer> = customerRepository.findByTrainerId(trainerId)

        return if (customerList.isEmpty()) ResponseEntity.notFound().build() else ResponseEntity.ok().body(customerList)
    }

    @Secured
    @PostMapping("customers/update_firebase_uid/{id}")
    fun updateCustomerFirebaseUidById(@PathVariable(value = "id") customerId: Long, @Valid @RequestBody firebaseUid: String)
            : ResponseEntity<Customer> {
        logger.info("Get customer id: $customerId uid: $firebaseUid")
        val returnCustomer: Customer = customerRepository.findById(customerId).orElse(null)
                ?: return ResponseEntity.notFound().build()

        returnCustomer.firebaseUid = firebaseUid
        return ResponseEntity.ok().body(customerRepository.save(returnCustomer))
    }

    @Secured
    @PostMapping("customers/update_password/{id}")
    fun updateCustomerPasswordById(@PathVariable(value = "id") customerId: Long, @Valid @RequestBody passwordB64: String)
            : ResponseEntity<Customer> {
        val returnCustomer: Customer = customerRepository.findById(customerId).orElse(null)
            ?: return ResponseEntity.notFound().build()

        returnCustomer.password = serviceBeans!!.passwordEncoder().encode(String(Base64.getDecoder().decode(passwordB64)))
        return ResponseEntity.ok().body(customerRepository.save(returnCustomer))
    }

    @Secured
    @PostMapping("customers/deactivate/{id}")
    fun deactivateCustomer(@PathVariable(value = "id") customerId: Long): ResponseEntity<Customer> {
        logger.info("Get customer id: $customerId")
        val returnCustomer: Customer = customerRepository.findById(customerId).orElse(null)
            ?: return ResponseEntity.notFound().build()

        returnCustomer.active = "D"
        return ResponseEntity.ok().body(customerRepository.save(returnCustomer))
    }

    @Secured
    @PostMapping("/customers/{id}")
    fun updateCustomerById(@PathVariable(value = "id") customerId: Long,
                           @Valid @RequestBody newCustomer: Customer,
                           @RequestHeader headers: HttpHeaders): ResponseEntity<Customer> {

        val returnCustomer: Customer = customerRepository.findById(customerId).orElse(null)
                ?: return ResponseEntity.notFound().build()

        val updatedCustomer = returnCustomer.copy()
        if (newCustomer.email != null) {
                updatedCustomer.email = newCustomer.email
            }
        if (newCustomer.name != null) {
                updatedCustomer.name = newCustomer.name
            }
        if (newCustomer.firstname != null) {
                updatedCustomer.firstname = newCustomer.firstname
            }
        if (newCustomer.goal != null) {
                updatedCustomer.goal = newCustomer.goal
            }
        if (newCustomer.dateChange != null) {
                updatedCustomer.dateChange = newCustomer.dateChange
            }
        if (newCustomer.dataPolicyAllowed != null) {
                updatedCustomer.dataPolicyAllowed = newCustomer.dataPolicyAllowed
            }
        if (newCustomer.sportId != null) {
            updatedCustomer.sportId = newCustomer.sportId
        }
        if (newCustomer.emailSubscription != null) {
            updatedCustomer.emailSubscription = newCustomer.emailSubscription
        }
        if (newCustomer.syncedDate != null) {
            updatedCustomer.syncedDate = newCustomer.syncedDate
        }
        if (newCustomer.trialDate != null) {
            updatedCustomer.trialDate = newCustomer.trialDate
        }
        if (newCustomer.firebaseRegToken != null) {
            updatedCustomer.firebaseRegToken = newCustomer.firebaseRegToken
        }
        if (newCustomer.lang != null) {
            updatedCustomer.lang = newCustomer.lang
        }
        updatedCustomer.sex = newCustomer.sex
        updatedCustomer.birthYear = newCustomer.birthYear
        updatedCustomer.fitnessLevel = newCustomer.fitnessLevel

        return ResponseEntity.ok().body(customerRepository.save(updatedCustomer))
    }

    @PostMapping("/club_registration")
    fun clubRegistration(@Valid @RequestBody json: String, @Value("\${firebase.key}") apiKey: java.lang.String): ResponseEntity<*> {

        val newUser: ClubUser = ClubUser().fromJson(json)

        if ( newUser.email.isEmpty()) {
            return ResponseEntity.badRequest().body("No Email")
        }

        val existingCustomer: Customer? = customerRepository.findByEmailAndActive(newUser.email, "Y")
        if (existingCustomer != null ) {
            return ResponseEntity.badRequest().body("Customer exists")
        }

        val current = LocalDateTime.now()
        val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS")
        val nowFormatted = current.format(formatter)

        val stringCharacters = ('0'..'z').toList().toTypedArray()
        val genPassword = (1..10).map { stringCharacters.random() }.joinToString("")

        val firebase = Firebase(apiKey)
        val signupResponse = firebase.signUp(newUser.email, genPassword)
            ?: return ResponseEntity.badRequest().body("Firebase exception ${firebase.error}")

        val savingCustomer = Customer()
        if ( serviceBeans == null ) {
            serviceBeans = ServiceBeans()
        }

        with (savingCustomer) {
            email = newUser.email
            password = serviceBeans!!.passwordEncoder().encode(genPassword)
            firebaseRegToken = signupResponse.idToken
            firebaseUid = signupResponse.localId
            dateAdd = nowFormatted
            firstname = newUser.firstname
            name = ""
            goal = newUser.goal
            fitnessLevel = newUser.fitnessLevel
            birthYear = newUser.birthYear
            sex = newUser.sex
        }

        val newCustomer = customerRepository.save(savingCustomer)

        if ( newUser.weight != 0.0 ) {
            val property = CustomerPropertyProperty()
            with (property) {
                propertyId = 1
                propertyValue = newUser.weight
                dateAdd= nowFormatted
                goal = false
                customer = newCustomer
            }
            newCustomer.properties.add(property)
        }
        if ( newUser.height != 0.0 ) {
            val property = CustomerPropertyProperty()
            with (property) {
                propertyId = 2
                propertyValue = newUser.height
                dateAdd = nowFormatted
                goal = false
                customer = newCustomer
            }
            newCustomer.properties.add(property)
        }

        val newMembershipId = newUser.membershipId
        if ( newMembershipId != 0L) {
            val membership = CustomerMembership()
            with(membership) {
                customer = newCustomer
                membershipId = newMembershipId
                startDate = nowFormatted
            }
            newCustomer.memberships.add(membership)
        }

        customerRepository.save(newCustomer)

        // create email link
        val activationLink = "https://club.workouttest.com/welcome/id=${signupResponse.idToken}"
        if ( emailTemplateService == null ) {
            emailTemplateService = EmailTemplateService()
        }
        val html = emailTemplateService!!.getEmailBody(newUser.firstname, activationLink, "registration_email")
        val subject = emailTemplateService!!.getSubject()

        // send email
        val email = Email()
        email.send(newUser.email, html, subject)

        return ResponseEntity.ok().body(newCustomer)
    }

    @PostMapping("/registration")
    fun registration(@Valid @RequestBody json: String): ResponseEntity<*> {
        val customer = Customer()

        val current = LocalDateTime.now()
        val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS")
        val nowFormatted = current.format(formatter)

        val newUser: User = User().fromJson(json)
        with (customer) {
            email = newUser.username
            password = serviceBeans!!.passwordEncoder().encode(newUser.password)
            firebaseUid = newUser.firebaseUid
            dateAdd = nowFormatted
            firstname = ""
            name = ""
            email = newUser.username
        }

        val returnCustomer: Customer? = customerRepository.findByEmailAndActive(newUser.username, "Y").let {
                if ( it == null) {
                    customerRepository.save(customer)
                } else {
                    null
                }
        }

        return if ( returnCustomer != null ) {
            ResponseEntity.ok().body(returnCustomer)
        } else {
            ResponseEntity.badRequest().body("Customer exists")
        }
    }



    @PostMapping("/login")
    fun login(@Valid @RequestBody json: String): ResponseEntity<*> {
        val customer = Customer()

        val newUser: User = User().fromJson(json)
        with (customer) {
            email = newUser.username
            password = newUser.password
        }
        val returnCustomer: Customer? = customerRepository.findByEmailAndActive(newUser.username, "Y").let {
            if ( it == null) {
                null
            } else {
                if (serviceBeans!!.passwordEncoder().matches(newUser.password, it.password)) {
                    it
                } else {
                    null
                }

            }
        }
        return if ( returnCustomer != null ) {
            ResponseEntity.ok().body(returnCustomer)
        } else {
            ResponseEntity.badRequest().body("Customer does not exist or the password is wrong")
        }
    }

    @PostMapping("/membership/{id}")
    fun newMembership(@PathVariable(value = "id") membershipId: Long, @Valid @RequestBody customer: Customer): ResponseEntity<Customer> {
        val returnCustomer: Customer = customerRepository.findById(customer.customerId).orElse(null)
            ?: return ResponseEntity.notFound().build()
        println("found customer ${returnCustomer.customerId}")

        val membership: Membership = membershipRepository.findByIdOrNull(membershipId)
            ?: return ResponseEntity.notFound().build()

        println("found membership $membershipId")

        val customerMembership = CustomerMembership()
        customerMembership.membershipId = membershipId
        customerMembership.customer = returnCustomer
        val now = LocalDateTime.now()
        val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")
        customerMembership.startDate = now.format(formatter)
        returnCustomer.memberships.add(customerMembership)
        customerRepository.save(returnCustomer)

        val savedCustomer: Customer = customerRepository.findById(customer.customerId).orElse(null)
            ?: return ResponseEntity.notFound().build()

        println("saved ${savedCustomer.customerId}")
        return ResponseEntity.ok(savedCustomer)
    }

    @PostMapping("/membership/cancel/{id}")
    fun cancelMembership(@PathVariable(value = "id") membershipId: Long, @Valid @RequestBody customer: Customer): ResponseEntity<Customer> {
        val returnCustomer: Customer = customerRepository.findById(customer.customerId).orElse(null)
            ?: return ResponseEntity.notFound().build()
        println("found customer ${returnCustomer.customerId}")

        val membership: Membership = membershipRepository.findByIdOrNull(membershipId)
            ?: return ResponseEntity.notFound().build()
        println("found membership $membershipId")

        var found = false
        var savedCustomer: Customer? = null
        for ( customerMembership in returnCustomer.memberships ) {
            if ( customerMembership.membershipId == membershipId) {
                val now = LocalDateTime.now()
                val formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
                customerMembership.endDate = now.format(formatter)
                savedCustomer = customerRepository.save(returnCustomer)
                found = true
            }
        }
        return if ( ! found ) {
            println("not found customer membership to cancel $membershipId")
            ResponseEntity.notFound().build()
        } else {
            ResponseEntity.ok(savedCustomer)
        }
    }
}